# Copyright (C) 2016-2025 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Test suite for module pyszn.

See http://pythontesting.net/framework/pytest/pytest-introduction/#fixtures
"""

from pyszn import __version__


def setup_module(module):
    print('setup_module({})'.format(module.__name__))


def teardown_module(module):
    print('teardown_module({})'.format(module.__name__))


def test_semantic_version():
    """
    Check that version follows the Semantic Versioning 2.0.0 specification.

        http://semver.org/
    """
    mayor, minor, rev = map(int, __version__.split('.'))

    assert mayor >= 0
    assert minor >= 0
    assert rev >= 0



# Copyright (C) 2015-2025 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
SZN parser module.

This module takes care of parsing a topology meta-description written in SZN.

This topology representation format allows to quickly specify simple topologies
that are only composed of simple nodes, ports and links between them. For a
more programmatic format consider the ``metadict`` format or using the
:class:`pynml.manager.ExtendedNMLManager` directly.

The format for the textual description of a topology is similar to Graphviz
syntax and allows to define nodes and ports with shared attributes and links
between two endpoints with shared attributes too.

::

    # Environment
    [kernel="3.13.0-77-generic"]

    # Nodes
    [type=switch attr1=1] sw1 sw2
    hs1

    # Ports
    [speed=1000] sw1:3 sw2:3

    # Links
    [linkattr1=20] sw1:a -- sw2:a
    [linkattr2=40] sw1:3 -- sw2:3

In the above example two nodes with the attributes ``type`` and ``attr1`` are
specified. Then a third node `hs1` with no particular attributes is defined.
Later, we specify some attributes (speed) for a couple of ports. In the same
way, a link between endpoints MAY have attributes. An endpoint is a combination
of a node and a port name.
"""

import logging
from re import split
from pathlib import Path
from copy import deepcopy
from textwrap import dedent
from traceback import format_exc
from collections import OrderedDict

from pyparsing import (
    Word, Literal, QuotedString,
    ParserElement, alphas, nums, alphanums,
    Group, OneOrMore, Optional,
    LineEnd, Suppress, LineStart, restOfLine, CaselessLiteral, Combine
)

log = logging.getLogger(__name__)


def naturalkey(element):
    """
    Return a tuple with the string and integer components of a given string.

    For example::

        naturalkey('node01') -> ('node', 1)

    :param str element: The string to decompose.

    :return: The string and integer components of the string.
    :rtype: tuple
    """
    return tuple(
        int(token) if token.isdigit() else token.lower()
        for token in split('([0-9]+)', element)
    )


class ParseException(Exception):
    """
    Custom exception thrown by the parser if it was unable to parse a SZN
    string.

    :var lineno: Line number of the parsing error.
    :var raw_line: Raw line that failed to be parsed.
    :var exc: Internal failure traceback of the parser.
    """

    def __init__(self, lineno, raw_line, exc):
        super(ParseException, self).__init__(
            'Unable to parse line #{}: "{}"'.format(
                lineno, raw_line
            )
        )
        self.lineno = lineno
        self.raw_line = raw_line
        self.exc = exc


def build_parser():
    """
    Build a pyparsing parser for our custom topology description language.

    :return: A pyparsing parser.
    :rtype: pyparsing.MatchFirst
    """
    ParserElement.set_default_whitespace_chars(' \t')
    nl = Suppress(LineEnd())
    inumber = Word(nums).set_parse_action(lambda toks: int(toks[0]))
    fnumber = (
        Combine(
            Optional('-') + Word(nums) + '.' + Word(nums)
            + Optional('E' | 'e' + Optional('-') + Word(nums))
        )
    ).set_parse_action(lambda toks: float(toks[0]))
    boolean = (
        CaselessLiteral('true') | CaselessLiteral('false')
    ).set_parse_action(lambda s, loc, toks: toks[0].casefold() == 'true')
    comment = Literal('#') + restOfLine + nl
    text = QuotedString('"')

    multiline_text = QuotedString('```', multiline=True)
    multiline_text.add_parse_action(lambda t: dedent(t[0]))

    identifier = Word(alphas, alphanums + '_')
    empty_line = LineStart() + LineEnd()
    item_list = (
        (text | fnumber | inumber | boolean)
        + Optional(Suppress(',')) + Optional(nl)
    )
    custom_list = (
        Suppress('(') + Optional(nl) + Group(OneOrMore(item_list))
        + Optional(nl) + Suppress(')')
    ).set_parse_action(lambda tok: tok.as_list())
    attribute = Group(
        identifier('key') + Suppress(Literal('='))
        + (
            custom_list | multiline_text | text | fnumber
            | inumber | boolean | identifier
        )('value')
        + Optional(nl)
    )
    attributes = (
        Suppress(Literal('['))
        + Optional(nl)
        + OneOrMore(attribute)
        + Suppress(Literal(']'))
    )

    node = Word(alphas, alphanums + '_' + '>')('node')

    port = Group(
        node + Suppress(Literal(':')) + (identifier | inumber)('port')
    )
    link = Group(
        port('endpoint_a') + Suppress(Literal('--')) + port('endpoint_b')
    )

    environment_spec = (
        attributes + nl
    ).set_results_name('env_spec', list_all_matches=True)
    nodes_spec = (
        Group(
            Optional(attributes)('attributes')
            + Group(OneOrMore(node))('nodes')
        ) + nl
    ).set_results_name('node_spec', list_all_matches=True)
    ports_spec = (
        Group(
            Optional(attributes)('attributes')
            + Group(OneOrMore(port))('ports')
        ) + nl
    ).set_results_name('port_spec', list_all_matches=True)
    link_spec = (
        Group(
            Optional(attributes)('attributes') + link('links')
        ) + nl
    ).set_results_name('link_spec', list_all_matches=True)

    statements = OneOrMore(
        comment
        | link_spec
        | ports_spec
        | nodes_spec
        | environment_spec
        | empty_line
    )
    return statements


def parse_txtmeta(txtmeta):
    """
    Parse a textual description of a topology and return a dictionary of it's
    elements.

    :param str txtmeta: The textual meta-description of the topology.
    :rtype: dict
    :return: Topology as dictionary.
    """
    statement = build_parser()
    data = {
        'environment': OrderedDict(),
        'nodes': [],
        'ports': [],
        'links': [],
    }

    parsed_result = statement.parse_string(txtmeta)

    # Process environment line
    if 'env_spec' in parsed_result:
        if len(parsed_result['env_spec']) > 1:
            raise Exception(
                'Multiple declaration of environment attributes: '
                '{}'.format(parsed_result['env_spec'][1])
            )
        for attr in parsed_result['env_spec'][0]:
            data['environment'][attr.key] = attr.value[0]

    # Process the links
    if 'link_spec' in parsed_result:
        for parsed in parsed_result['link_spec']:
            link = parsed[0].links
            attrs = OrderedDict()
            if "attributes" in parsed[0]:
                for attr in parsed[0].attributes:
                    attrs[attr.key] = attr.value[0]
            data['links'].append({
                'endpoints': (
                    (str(link.endpoint_a.node), str(link.endpoint_a.port)),
                    (str(link.endpoint_b.node), str(link.endpoint_b.port)),
                ),
                'attributes': attrs,
            })

            data['nodes'].append({
                'nodes': [
                    str(link.endpoint_a.node), str(link.endpoint_b.node)
                ],
                'attributes': OrderedDict(),
            })
            data['ports'].append({
                'ports': [
                    (str(link.endpoint_a.node), str(link.endpoint_a.port)),
                    (str(link.endpoint_b.node), str(link.endpoint_b.port))
                ],
                'attributes': OrderedDict(),
            })

    # Process the ports
    if 'port_spec' in parsed_result:
        for parsed in parsed_result['port_spec']:
            ports = parsed[0].ports
            attrs = OrderedDict()
            if "attributes" in parsed[0]:
                for attr in parsed[0].attributes:
                    attrs[attr.key] = attr.value[0]
            data['ports'].append({
                'ports': [
                    (str(port.node), str(port.port)) for port in ports
                ],
                'attributes': attrs,
            })
            data['nodes'].append({
                'nodes': [str(port.node) for port in ports],
                'attributes': OrderedDict(),
            })

    # Process the nodes
    if 'node_spec' in parsed_result:
        for parsed in parsed_result['node_spec']:
            nodes = parsed[0].nodes
            attrs = OrderedDict()
            if 'attributes' in parsed[0]:
                for attr in parsed[0].attributes:
                    attrs[attr.key] = attr.value[0]
            data['nodes'].append({
                'nodes': [str(node) for node in nodes],
                'attributes': attrs,
            })

    # Remove duplicate data created implicitly
    temp_nodes = OrderedDict()
    for node_list in data['nodes']:
        for node in node_list['nodes']:
            if node not in temp_nodes.keys():
                temp_nodes[node] = {
                    'attributes': deepcopy(node_list['attributes']),
                    'nodes': [node],
                    'subnodes': []
                }
            else:
                temp_nodes[node]['attributes'].update(node_list['attributes'])

            depth = node.count('>')
            parent = None if depth == 0 else node.rsplit('>', 1)[-2]
            temp_nodes[node]['parent'] = parent
            temp_nodes[node]['depth'] = depth

    for node, spec in sorted(
        temp_nodes.items(), key=lambda t: t[1]['depth'], reverse=True
    ):
        parent = spec['parent']
        if parent is not None:
            parent_node = temp_nodes[parent]
            siblings = parent_node['subnodes']
            siblings.append(node)

        del spec['depth']

    temp_ports = OrderedDict()
    for port_list in data['ports']:
        for port in port_list['ports']:
            if port not in temp_ports.keys():
                temp_ports[port] = deepcopy(port_list['attributes'])
            else:
                temp_ports[port].update(port_list['attributes'])

    data['nodes'] = [
        {
            'nodes': [k],
            'attributes': v['attributes'],
            'subnodes': v['subnodes'],
            'parent': v['parent'],
        }
        # Sort by node name
        for k, v in sorted(
            temp_nodes.items(),
            key=lambda e: naturalkey(e[0]),
        )
    ]

    data['ports'] = [
        {'ports': [k], 'attributes': v}
        # Sort by node name and port name ('node01', 'port01')
        for k, v in sorted(
            temp_ports.items(),
            key=lambda e: tuple(
                naturalkey(part)
                for part in e[0]
            ),
        )
    ]

    data['links'] = sorted(
        data['links'],
        # Sort by endpoints (('node01', 'port01'), ('node02', 'port02'))
        key=lambda e: tuple(
            tuple(
                naturalkey(subpart)
                for subpart in part
            )
            for part in e['endpoints']
        )
    )

    return data


def find_topology_in_python(filename, szn_dir=None, encoding='utf-8'):
    """
    Find the topology definition inside a Python file.

    This function searches for a variable named TOPOLOGY or TOPOLOGY_ID inside
    the given Python file. If TOPOLOGY is found, it's value is returned
    directly. If TOPOLOGY_ID is found instead, the function searches for a file
    named <TOPOLOGY_ID>.szn inside the given szn_dir list of paths. If found,
    the content of that file is returned.

    This helper functions build a AST tree a grabs the variable from it. Thus,
    the Python code isn't executed.

    :param Path filename: Path to file to search for TOPOLOGY or TOPOLOGY_ID.
    :param list szn_dir: List of paths to directories where \\*.szn files are
     located.
    :param str encoding: Encoding used to read the files.

    :return: The value of the TOPOLOGY variable if found, None otherwise.
    :rtype: str or None
    """
    import ast

    # Backward compatibility for string paths
    if not isinstance(filename, Path):
        filename = Path(filename)
    if szn_dir is not None:
        szn_dir = [
            Path(path)
            if not isinstance(path, Path)
            else path
            for path in szn_dir
        ]

    try:
        # Parse the Python file to extract the TOPOLOGY or TOPOLOGY_ID variable
        tree = ast.parse(filename.read_text(encoding=encoding))

        # Iterate over the AST nodes to find our variable
        for node in ast.iter_child_nodes(tree):

            # Check if the node is an assignment, if not, skip it
            if not isinstance(node, ast.Assign):
                continue

            # Get the list of assignment targets
            targets = {targets.id for targets in node.targets}

            # Check if we have a TOPOLOGY variable
            if 'TOPOLOGY' in targets:

                # This used to be node.value.s in Python 3.7 and earlier
                # But was deprecated in favor of node.value.value in
                # Python 3.8+
                return node.value.value

            # Check if we have a TOPOLOGY_ID variable
            elif 'TOPOLOGY_ID' in targets:

                if not szn_dir:
                    raise RuntimeError(
                        'Found a TOPOLOGY_ID, but no SZN search '
                        'path was defined'
                    )

                # Grab the reference to the external file
                topology_id = node.value.value

                # Iterate over the search paths to find the referenced file
                # Iteration is done following the explicit order of the list
                # So the user can prioritize paths if needed
                for search_path in szn_dir:

                    # According to Python documentation, Path.glob is not
                    # deterministic, so we need to sort the results to have a
                    # predictable behavior
                    for filename in sorted(
                        search_path.glob(f'{topology_id}.szn'),
                        # Length first (shortest path)
                        # Natural sorting tie-breaker
                        key=lambda path: (
                            len(str(path)),
                            naturalkey(str(path)),
                        )
                    ):
                        return filename.read_text(encoding=encoding)

                raise FileNotFoundError(
                    f'Topology file with ID {topology_id} could not be found'
                )

    except Exception:
        log.error(format_exc())
    return None


__all__ = [
    'ParseException',
    'parse_txtmeta',
    'find_topology_in_python'
]



# Copyright (C) 2015-2025 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Test suite for module pyszn.injection.
"""

from os.path import join
from shutil import rmtree
from collections import OrderedDict

from deepdiff import DeepDiff

from pyszn.injection import parse_attribute_injection


TOPOLOGY_TO_BUILD = """
# Nodes
[shell=vtysh name="Switch 9"] sw9
[shell=vtysh name="Switch 2"] sw2
[type=switch name="Switch 8"] sw8
[type=host name="Host 3"] hs3
[type=host name="Host 2"] hs2
# Links
hs3:1 -- sw9:1
hs2:1 -- sw2:1
[attr1=value1] sw9:2 -- sw2:2
"""

TOPOLOGY_MATCH_0 = """
# Nodes
[shell=vtysh name="Switch 1"] sw1
[shell=vtysh name="Switch 2"] sw2
[type=switch name="Switch 8"] sw8
[type=host name="Host 1"] hs1
[type=host name="Host 2"] hs2
# Links
hs1:1 -- sw1:1
hs2:1 -- sw2:1
[attr1=1] sw1:2 -- sw2:2
"""

TOPOLOGY_MATCH_1 = """
# Nodes
[shell=vtysh name="Switch 1"] sw1
[shell=vtysh name="Switch 2"] sw2
[type=host name="Host 1"] hs1
[type=host name="Host 2"] hs2
# Links
hs1:1 -- sw1:1
hs2:1 -- sw2:1
[attr1=1] sw1:2 -- sw2:2
"""

TOPOLOGY_MATCH_2 = """
# Nodes
[shell=vtysh name="Switch 1"] sw4
[shell=vtysh name="Switch 2"] sw5
"""

TOPOLOGY_MATCH_3 = """
# Nodes
[shell=vtysh name="Switch 1"] sw6
[shell=vtysh name="Switch 2"] sw7
[type=host name="Host 1"] hs3
[type=host name="Host 2"] hs4
# Links
hs1:1 -- sw1:1
hs2:1 -- sw2:1
[attr1=1] sw1:2 -- sw2:2
"""

TOPOLOGY_MATCHES = {
    '0': TOPOLOGY_MATCH_0,
    '2': TOPOLOGY_MATCH_2
}

TOPOLOGY_MATCHES_FOLDER = {
    '1': TOPOLOGY_MATCH_1,
    '3': TOPOLOGY_MATCH_3
}

INJECTION_FILE = """
[
    {{
        "files": ["test_topology_match_0.py"],
        "modifiers": [
             {{
                "links": ["hs1:1 -- sw1:1", "test_attr=test"],
                "attributes": {{
                    "link_attr": "link_value"
                }}
           }},
           {{
                "ports": ["hs1:1", "test_attr=test"],
                "attributes": {{
                    "port_attr": "port_value"
                }}
            }},
            {{
                "nodes": ["sw1"],
                "attributes": {{
                    "image": "image_for_sw1",
                    "hardware": "hardware_for_sw1"
                }}
            }},
            {{
                "nodes": ["sw2"],
                "attributes": {{
                    "image": "image_for_sw2",
                    "shell": "vtysh",
                    "name":  "new_name"
                }}
            }},
            {{
                "nodes": ["type=switch"],
                "attributes": {{
                    "chassis": "chassis_for_sw8"
                }}
            }}
        ]
    }},
    {{
        "files": [
            "test_topology_match_0.py",
            "test_topology_match_1.py"
        ],
        "modifiers": [
            {{
                "nodes": ["sw1", "type=host", "sw3"],
                "attributes": {{
                    "image": "image_for_sw1_sw3_hs1_hs2",
                    "hardware": "hardware_for_sw1_sw3_hs1_hs2"
                }}
            }},
            {{
                "nodes": ["sw4"],
                "attributes": {{
                    "image": "image_for_sw4"
                }}
            }}
        ]
    }},
    {{
        "files": ["test_topology_match_2.py"],
        "modifiers": [
            {{
                "nodes": ["*"],
                "attributes": {{
                    "image": "image_for_sw4_sw5"
                }}
            }}
        ]
    }},
    {{
        "files": ["*"],
        "modifiers": [
            {{
                "nodes": ["hs*"],
                "attributes": {{
                    "image": "image_for_all_hosts"
                }}
            }},
            {{
                "nodes": ["sw6", "sw7"],
                "attributes": {{
                    "image": "image_for_sw6_sw7"
                }}
            }}
        ]
    }}
]
"""

EXPECTED_PARSED_INJECTION_FILE = OrderedDict([
    (
        '{search_path}/test_topology_match_0.py',
        {
            'environment': {},
            'ports': OrderedDict([
                (('hs1', '1'), {'port_attr': 'port_value'})
            ]),
            'links': OrderedDict([(
                (('hs1', '1'), ('sw1', '1')),
                {
                    'link_attr': 'link_value',
                }
            )]),
            'nodes': OrderedDict([
                (
                    'sw1', {
                        'image': 'image_for_sw1_sw3_hs1_hs2',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ), (
                    'sw2', {
                        'image': 'image_for_sw2',
                        'shell': 'vtysh',
                        'name': 'new_name',
                    }
                ), (
                    'sw8', {
                        'chassis': 'chassis_for_sw8'
                    }
                ), (
                    'hs1', {
                        'image': 'image_for_all_hosts',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ), (
                    'hs2', {
                        'image': 'image_for_all_hosts',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ),
            ])}
    ), (
        '{search_path}/subfolder/test_topology_match_1.py',
        {
            'environment': {},
            'ports': OrderedDict(),
            'links': OrderedDict(),
            'nodes': OrderedDict([
                (
                    'sw1', {
                        'image': 'image_for_sw1_sw3_hs1_hs2',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ), (
                    'hs1', {
                        'image': 'image_for_all_hosts',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ), (
                    'hs2', {
                        'image': 'image_for_all_hosts',
                        'hardware': 'hardware_for_sw1_sw3_hs1_hs2',
                    }
                ),
            ])
        }
    ), (
        '{search_path}/test_topology_match_2.py',
        {
            'environment': {},
            'ports': OrderedDict(),
            'links': OrderedDict(),
            'nodes': OrderedDict([
                (
                    'sw4', {
                        'image': 'image_for_sw4_sw5',
                    }
                ), (
                    'sw5', {
                        'image': 'image_for_sw4_sw5',
                    }
                ),
            ])
        }
    ), (
        '{search_path}/subfolder/test_topology_match_3.py',
        {
            'environment': {},
            'ports': OrderedDict(),
            'links': OrderedDict(),
            'nodes': OrderedDict([
                (
                    'hs1', {
                        'image': 'image_for_all_hosts',
                    }
                ), (
                    'hs2', {
                        'image': 'image_for_all_hosts',
                    }
                ), (
                    'hs3', {
                        'image': 'image_for_all_hosts',
                    }
                ), (
                    'hs4', {
                        'image': 'image_for_all_hosts',
                    }
                ), (
                    'sw6', {
                        'image': 'image_for_sw6_sw7',
                    }
                ), (
                    'sw7', {
                        'image': 'image_for_sw6_sw7',
                    }
                ),
            ])
        }
    )]
)


def test_attribute_injection(tmpdir):
    """
    Test the configuration file is parsed correctly.
    """
    workdir = str(tmpdir)
    search_path = str(tmpdir.mkdir('test'))
    subfolder = str(tmpdir.mkdir('test/subfolder'))

    try:
        # Write matching topologies
        for basepath, matches in (
                (search_path, TOPOLOGY_MATCHES),
                (subfolder, TOPOLOGY_MATCHES_FOLDER)):
            for count, content in matches.items():
                output_filename = join(
                    basepath, 'test_topology_match_{}.py'.format(count)
                )
                with open(output_filename, 'w') as fd:
                    fd.write('TOPOLOGY = """\n')
                    fd.write(content)
                    fd.write('"""')

        # Write the attributes injection file
        injection_path = join(workdir, 'attributes_injection.json')
        with open(injection_path, 'w') as fd:
            fd.write(INJECTION_FILE.format(search_path=search_path))

        # Change keys of the expected parsed injection file
        expected = OrderedDict()
        for key, value in EXPECTED_PARSED_INJECTION_FILE.items():
            expected[key.format(search_path=search_path)] = value

        # Actually parse the injection file
        actual = parse_attribute_injection(
            injection_path, search_paths=[search_path]
        )

        # Compare the actual and the expected
        differences = DeepDiff(actual, expected)
        assert not differences

    finally:
        try:
            rmtree(workdir)
        except Exception:
            pass



# Copyright (C) 2015-2025 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Test suite for module pyszn.parser.
"""

from textwrap import dedent
from collections import OrderedDict
from logging import getLogger

from deepdiff import DeepDiff


from pyszn.parser import parse_txtmeta


log = getLogger(__name__)


def test_basic_parse():
    """
    Tests parsing of a complete SZN
    """

    topology = """
    # Environment
    [virtual=none awesomeness=medium]

    # Nodes
    [shell=vtysh] sw1 sw2
    [type=host] hs1
    hs2

    # Links
    sw1:1 -- hs1:1
    [attr1=2.1e2 attr2=-2.7e-1] sw1:a -- hs1:a
    [attr1=1 attr2="lorem ipsum" attr3=(1, 3.0, "B")] sw1:4 -- hs2:a
    """

    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(
            [('virtual', 'none'), ('awesomeness', 'medium')]
        ),
        'nodes': [
            {
                'attributes': OrderedDict([('type', 'host')]),
                'nodes': ['hs1'],
                'subnodes': [],
                'parent': None,
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['hs2'],
                'subnodes': [],
                'parent': None,
            },
            {
                'attributes': OrderedDict([('shell', 'vtysh')]),
                'nodes': ['sw1'],
                'subnodes': [],
                'parent': None,
            },
            {
                'attributes': OrderedDict([('shell', 'vtysh')]),
                'nodes': ['sw2'],
                'subnodes': [],
                'parent': None,
            },
        ],
        'ports': [
            {
                'ports': [('hs1', '1')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('hs1', 'a')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('hs2', 'a')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('sw1', '1')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('sw1', '4')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('sw1', 'a')],
                'attributes': OrderedDict()
            },
        ],
        'links': [
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw1', '1'), ('hs1', '1'))
            },
            {
                'attributes': OrderedDict(
                    [
                        ('attr1', 1), ('attr2', 'lorem ipsum'),
                        ('attr3', [1, 3.0, 'B'])
                    ]
                ),
                'endpoints': (('sw1', '4'), ('hs2', 'a'))
            },
            {
                'attributes': OrderedDict(
                    [
                        ('attr1', 210.0),
                        ('attr2', -0.27)
                    ]
                ),
                'endpoints': (('sw1', 'a'), ('hs1', 'a'))
            },
        ]
    }

    assert not DeepDiff(actual, expected)


def test_autonode():
    """
    Test the automatic creation of implicit nodes
    """

    topology = """
    sw1:port1
    """

    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(),
        'nodes': [{
            'attributes': OrderedDict(),
            'nodes': ['sw1'],
            'subnodes': [],
            'parent': None,
        }],
        'ports': [{'ports': [('sw1', 'port1')], 'attributes': OrderedDict()}],
        'links': []
    }

    assert not DeepDiff(actual, expected)


def test_multiline_list():
    """
    Test the support for multiline list attributes
    """
    topology = """
    # Environment
    [
        virtual=none
        awesomeness=medium
        float=1.0
        list=(
            1,
            3.14,
            True
        )
    ]
    """
    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(
            [
                ('virtual', 'none'), ('awesomeness', 'medium'), ('float', 1.0),
                ('list', [1, 3.14, True])
            ]
        ),
        'nodes': [],
        'ports': [],
        'links': []
    }

    assert not DeepDiff(actual, expected)


def test_multiline_text():
    """
    Test the support for multiline text attributes
    """
    topology = '''
    # Environment
    [
        virtual=none
        awesomeness=medium
        float=1.0
        multiline_text=```
            Buenos Aires
            se ve tan susceptible
            Es el destino de furia, es
                lo que en sus caras persiste
        ```
    ]
    '''
    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(
            [
                ('virtual', 'none'),
                ('awesomeness', 'medium'),
                ('float', 1.0),

                # Multiline string attribute values are parsed using
                # textwrap.dedent()
                (
                    'multiline_text',
                    dedent("""
                        Buenos Aires
                        se ve tan susceptible
                        Es el destino de furia, es
                            lo que en sus caras persiste
                    """)
                )
            ]
        ),
        'nodes': [],
        'ports': [],
        'links': []
    }

    assert not DeepDiff(actual, expected)


def test_attributes():
    """
    Test that attributes should just be added to the nodes on the same line
    """
    topology = """
    [attr=value] hs1 hs3
    hs2
    """

    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(),
        'nodes': [
            {
                'attributes': OrderedDict([('attr', 'value')]),
                'nodes': ['hs1'],
                'subnodes': [],
                'parent': None,
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['hs2'],
                'subnodes': [],
                'parent': None,
            },
            {
                'attributes': OrderedDict([('attr', 'value')]),
                'nodes': ['hs3'],
                'subnodes': [],
                'parent': None,
            },
        ],
        'ports': [],
        'links': []
    }

    assert not DeepDiff(actual, expected)


def test_single():
    """
    Test that a single line string (no new lines '\\n') is parsed
    """
    topology = """[attr=value] hs1"""

    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(),
        'nodes': [
            {
                'attributes': OrderedDict([('attr', 'value')]),
                'nodes': ['hs1'],
                'subnodes': [],
                'parent': None,
            },
        ],
        'ports': [],
        'links': []
    }

    assert not DeepDiff(actual, expected)


def test_subnodes():
    """
    Test the capability of specifying subnodes.
    """
    topology = """
    # Level 0  nodes
    [color=red] switch1
    [color=blue] switch2

    # Level 1 node
    [flavor=vanilla] switch1>card1

    # Level 2 node
    [flavor="blue-berry"] switch1>card1>subcard1

    # Links
    # Link from a level 0 node to a level 0 node
    [link_mood=happy] switch2:port1 -- switch1:port1

    # Link from a level 0 node to a level 1 node
    switch2:port2 -- switch1>card1:port1

    # Link from a level 0 node to a level 2 node
    switch2:port3 -- switch1>card1>subcard1:port1

    # Ports
    # Level 0 port
    [speed_gbps=10] switch1:port1

    # Level 1 port
    [speed_gbps=20] switch1>card1:port1

    # Level 2 port
    [speed_gbps=30] switch1>card1>subcard1:port1
    """
    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(),

        # Nodes are sorted alphabetically by node name
        'nodes': [
            {
                'attributes': OrderedDict([('color', 'red')]),
                'nodes': ['switch1'],
                'subnodes': ['switch1>card1'],
                'parent': None,
            },
            {
                'attributes': OrderedDict([('flavor', 'vanilla')]),
                'nodes': ['switch1>card1'],
                'subnodes': ['switch1>card1>subcard1'],
                'parent': 'switch1',
            },
            {
                'attributes': OrderedDict([('flavor', 'blue-berry')]),
                'nodes': ['switch1>card1>subcard1'],
                'subnodes': [],
                'parent': 'switch1>card1',
            },
            {
                'attributes': OrderedDict([('color', 'blue')]),
                'nodes': ['switch2'],
                'subnodes': [],
                'parent': None,
            },

        ],

        # Ports are sorted alphabetically by the ports field
        'ports': [
            {
                'ports': [('switch1', 'port1')],
                'attributes': OrderedDict([('speed_gbps', 10)])
            },
            {
                'ports': [('switch1>card1', 'port1')],
                'attributes': OrderedDict([('speed_gbps', 20)])
            },
            {
                'ports': [('switch1>card1>subcard1', 'port1')],
                'attributes': OrderedDict([('speed_gbps', 30)])
            },
            {
                'ports': [('switch2', 'port1')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('switch2', 'port2')],
                'attributes': OrderedDict()
            },
            {
                'ports': [('switch2', 'port3')],
                'attributes': OrderedDict()
            },
        ],

        # Links are sorted alphabetically by the endpoints field
        'links': [
            {
                'attributes': OrderedDict(
                    [
                        ('link_mood', 'happy'),
                    ]
                ),
                'endpoints': (('switch2', 'port1'), ('switch1', 'port1'))
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('switch2', 'port2'), ('switch1>card1', 'port1'))
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (
                    ('switch2', 'port3'), ('switch1>card1>subcard1', 'port1')
                )
            },
        ]
    }
    assert not DeepDiff(actual, expected)


def test_natural_order():
    """
    Tests order of node, ports and links
    """

    topology = """
    # Nodes
    sw1
    sw2
    sw10
    hs1
    hw2
    hs10

    # Links
    sw1:1 -- hs1:1
    sw1:2 -- hs2:1
    sw1:10 -- hs10:1

    sw2:1 -- hs1:10
    sw2:2 -- hs2:11
    sw2:10 -- hs10:12
    """

    actual = parse_txtmeta(topology)

    expected = {
        'environment': OrderedDict(),
        'links': [
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw1', '1'), ('hs1', '1'))
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw1', '2'), ('hs2', '1'))
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw1', '10'), ('hs10', '1')),
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw2', '1'), ('hs1', '10')),
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw2', '2'), ('hs2', '11')),
            },
            {
                'attributes': OrderedDict(),
                'endpoints': (('sw2', '10'), ('hs10', '12')),
            },
        ],
        'nodes': [
            {
                'attributes': OrderedDict(),
                'nodes': ['hs1'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['hs2'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['hs10'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['hw2'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['sw1'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['sw2'],
                'parent': None,
                'subnodes': [],
            },
            {
                'attributes': OrderedDict(),
                'nodes': ['sw10'],
                'parent': None,
                'subnodes': [],
            },
        ],
        'ports': [
            {'attributes': OrderedDict(), 'ports': [('hs1', '1')]},
            {'attributes': OrderedDict(), 'ports': [('hs1', '10')]},
            {'attributes': OrderedDict(), 'ports': [('hs2', '1')]},
            {'attributes': OrderedDict(), 'ports': [('hs2', '11')]},
            {'attributes': OrderedDict(), 'ports': [('hs10', '1')]},
            {'attributes': OrderedDict(), 'ports': [('hs10', '12')]},
            {'attributes': OrderedDict(), 'ports': [('sw1', '1')]},
            {'attributes': OrderedDict(), 'ports': [('sw1', '2')]},
            {'attributes': OrderedDict(), 'ports': [('sw1', '10')]},
            {'attributes': OrderedDict(), 'ports': [('sw2', '1')]},
            {'attributes': OrderedDict(), 'ports': [('sw2', '2')]},
            {'attributes': OrderedDict(), 'ports': [('sw2', '10')]},
        ],
    }

    assert not DeepDiff(actual, expected)



# Copyright (C) 2015-2025 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Parser attributes injection files.

This module defines a set of functions that implement logic to parse a
attributes injection file and extract from it the injection specification that
allows to modify attributes of entities in different topologies.
"""

from re import match
from glob import glob
from json import loads
from os import getcwd, walk
from fnmatch import fnmatch
from logging import getLogger
from traceback import format_exc
from collections import OrderedDict
from os.path import isabs, join, abspath, isfile, basename

from pyszn.parser import parse_txtmeta, find_topology_in_python


log = getLogger(__name__)


def parse_attribute_injection(
    injection_file,
    search_paths=None,
    ignored_paths=None,
    szn_dir=None,
):
    """
    Parses a attributes injection file into an attribute injection dictionary.

    An attribute injection file is a JSON file that specifies a list of
    injection specifications. Each specification allows to list the files to
    modify and the modifiers to apply. Each modifier specify what entities
    (devices, ports or links) will be modified and which attributes will be
    injected. The file support willcards and attributes matching.

    The file format is as follows:

    ::

        [
            {
                "files": ["pathto/*", "another/foo*.py"],
                "modifiers": [
                    {
                        "nodes": ["sw1", "type=host", "sw3"],
                        "attributes": {
                            "image": "image_for_sw1_sw3_hs1_hs2",
                            "hardware": "hardware_for_sw1_sw3_hs1_hs2"
                        }
                    },
                    {
                        "nodes": ["sw4"],
                        "attributes": {
                            "image": "image_for_sw4"
                        }
                    },
                    ... # More modifiers
                ]
            },
            ... # More injection specifications
        ]

    :param str injection_file: Path for the attribute injection file.
    :param list search_paths: Paths to search for files when the file match is
     relative in the injection file.
     If ``None`` (the default), the current working directory is used.
    :param list ignored_paths: Lists of path pattern to ignore of the search
     paths. Uses fnmatch to filter them out.
    :param list szn_dir: List of paths to directories where \\*.szn files
     are located.
    :return: An ordered dictionary with the attributes to inject of the form:

     ::

        {
            '/abs/path/to/suite.py': {
                'sw1': {
                    'image': 'image_for_sw1_sw3_hs1_hs2',
                    'hardware': 'hardware_for_sw1_sw3_hs1_hs2'
                },
                'sw3': {
                    'image': 'image_for_sw1_sw3_hs1_hs2',
                    'hardware': 'hardware_for_sw1_sw3_hs1_hs2'
                }
            }
        }

    :rtype: `collections.OrderedDict`
    """
    # Define search paths
    if search_paths is None:
        search_paths = [abspath(getcwd())]
    log.debug('Injection search paths: {}'.format(search_paths))

    ignored_paths = ignored_paths or []
    log.debug('Injection will ignore: {}'.format(ignored_paths))

    # Expand search paths recursively to include all subfolders
    paths_to_expand = list(search_paths)
    for root in paths_to_expand:
        children = _subfolders(root, ignored_paths)
        search_paths.extend(children)

    # Make search paths unique
    uniques = []
    for path in search_paths:
        if path not in uniques:
            uniques.append(path)
    search_paths = uniques
    log.debug('Expanded injection search paths: {}'.format(search_paths))

    # Read injection file
    with open(injection_file) as fd:
        injection_spec = loads(fd.read())

    # Cache the parsed topologies to not parse them again and again
    # Key is a topology string and return is a dictionary of the parsed
    # topology
    parsed_topologies_cache = {}

    result = OrderedDict()

    # Iterate all specifications, expand files and fill return dictionary
    for spec in injection_spec:
        for filename in _expand_files(spec['files'], search_paths):

            if filename not in result:
                result[filename] = {}
                result[filename]['environment'] = {}
                result[filename]['nodes'] = OrderedDict()
                result[filename]['ports'] = OrderedDict()
                result[filename]['links'] = OrderedDict()

            # Environment attributes are parsed from the spec
            try:
                for attribute, value in spec['environment'].items():
                    result[filename]['environment'][attribute] = value
            except KeyError:
                pass

            topology = _load_topo(filename, szn_dir)

            # Check if topology exists in cache, otherwise parse it and save it
            if topology in parsed_topologies_cache:
                parsed_topology = parsed_topologies_cache[topology]
            else:
                try:
                    parsed_topology = parse_txtmeta(topology)
                    parsed_topologies_cache[topology] = parsed_topology
                except Exception:
                    log.error(
                        (
                            'Skipping node expansion for attribute injection '
                            'in filename {} in the lookup path as SZN format '
                            'parsing failed.'
                        ).format(filename))
                    log.debug(format_exc())
                    continue

            # Each specification have several "modifiers" associated to it.
            # Those modifiers hold the nodes whose attributes are to be
            # modified.
            for modifier in spec['modifiers']:

                # Nodes
                if 'nodes' in modifier:
                    for node in _expand_nodes(
                        parsed_topology, modifier['nodes']
                    ):
                        if node not in result[filename]['nodes']:
                            result[filename]['nodes'][node] = {}

                        for attribute, value in modifier[
                                'attributes'].items():
                            result[filename]['nodes'][node][
                                attribute] = value

                # Ports
                if 'ports' in modifier:
                    for port in _expand_ports(
                        parsed_topology,
                        _port_str_to_tuple(modifier['ports'])
                    ):
                        if port not in result[filename]['ports']:
                            result[filename]['ports'][port] = {}

                        for attribute, value in modifier[
                                'attributes'].items():
                            result[filename]['ports'][port][
                                attribute] = value

                # Links
                if 'links' in modifier:
                    for link in _expand_links(
                        parsed_topology,
                        _link_str_to_tuple(modifier['links'])
                    ):
                        if link not in result[filename]['links']:
                            result[filename]['links'][link] = {}

                        for attribute, value in modifier[
                                'attributes'].items():
                            result[filename]['links'][link][
                                attribute] = value

    log.debug('Attribute injection interpreted dictionary:')
    log.debug(result)

    return result


def _subfolders(search_path, ignored_paths):
    result = []
    for root, dirs, files in walk(search_path, topdown=True, followlinks=True):
        dirs[:] = [
            d for d in dirs
            if (
                # Ignore hidden folders
                not d.startswith('.') and
                # Ignore folders matching any ignored path pattern
                not any(fnmatch(d, pattern) for pattern in ignored_paths)
            )
        ]
        result.extend([join(root, directory) for directory in dirs])
    return result


def _port_str_to_tuple(port_str_list):
    result = []
    for port in port_str_list:
        if not match(r'(.*)=(.*)', port):
            port_host, port_num = port.split(':')
            result.append((port_host, port_num))
        else:
            result.append(port)
    return result


def _link_str_to_tuple(link_str_list):
    result = []
    for link in link_str_list:
        if not match(r'(.*)=(.*)', link):
            link = link.replace(' ', '')
            endpoint1, endpoint2 = link.split('--')
            link1_host, link1_port = endpoint1.split(':')
            link2_host, link2_port = endpoint2.split(':')
            result.append(((link1_host, link1_port), (link2_host, link2_port)))
        else:
            result.append(link)
    return result


def _load_topo(filename, szn_dir):
    """
    Load a topology string from a file.

    :param str filename: A filename to check for a topology string.
    :param list szn_dir: List of paths to directories where *.szn files are
     located.

    :return: A topology string if found, otherwise None.
    :rtype: str or None
    """

    # Find the topology string in the Python file
    if filename.endswith('.py'):
        topology = find_topology_in_python(filename, szn_dir=szn_dir)
        if topology is None:
            log.warning(
                (
                    'Skipping node expansion for attribute injection in '
                    'filename {} in the lookup path as it does not contain '
                    ' a TOPOLOGY or TOPOLOGY_ID definition.'
                ).format(filename))
    else:
        # *.szn files just contain the topology string directly
        with open(filename, 'r') as fd:
            topology = fd.read().strip()

    return topology


def _expand_files(files_definitions, search_paths):
    """
    Expands a list of files definitions into the matching files paths.

    A file definition is a string that can match none, one or more files
    (by using wildcards). It can be an absolute path, or a relative path from
    the search paths. For example:

    ::

        '/abs/path/to/my*_thing.py'
        'myfile.szn'
        'relative/test_*.py'

    :param list files_definitions: A list of files definitions.
    :param str search_paths: Paths to search for files when the file definition
     is relative.
    :return: A list of files paths.
    """

    expanded_files = []

    for file_definition in files_definitions:

        # File definitions to look for
        lookups = []

        # Determine if suite must be located in suites search path
        if isabs(file_definition):
            lookups.append(file_definition)
        else:
            for search_path in search_paths:
                lookups.append(join(search_path, file_definition))

        # Find all file matches for the suite definition
        matches = []
        for lookup in lookups:
            for filepath in glob(lookup):
                filename = basename(filepath)

                if filepath in expanded_files or not isfile(filepath):
                    continue

                if fnmatch(filename, 'test_*.py') or \
                        fnmatch(filename, '*.szn'):
                    matches.append(filepath)

        expanded_files.extend(matches)

    return expanded_files


def _expand_nodes(parsed_topology, nodes_definitions):
    """
    Expands a list of node definitions into the matching node names.

    A node definition is a string that can match none, one or more nodes
    (by using wildcards). It can be an expression for node name matching, or
    for matching all nodes that have an specific attribute value. For example:

    ::

        'nodea'
        'hs*'
        'type=host'

    :param dict parsed_topology: A dictionary of a fully parsed topology.
    :param list nodes_definitions: A list of node definitions.

    :return: A list of matching nodes.
    """

    expanded_nodes = []

    for node_definition in nodes_definitions:

        # Check if definition is for attribute matching
        if match(r'(.*)=(.*)', node_definition):
            expanded_nodes.extend(
                _match_by_attr(node_definition, parsed_topology, 'nodes'))
            continue

        # The definition is not attribute matching, but name matching
        for nodes_group in parsed_topology['nodes']:
            for node in nodes_group['nodes']:
                if fnmatch(node, node_definition) and \
                        node not in expanded_nodes:
                    expanded_nodes.append(node)

    return expanded_nodes


def _match_by_attr(regex, parsed_topology, kind):
    """
    Searches a parsed topology for attributes that match a regex.

    :param str regex: The regex to match the attribute, must of the form
     (attribute_regex)=(value_regex).
    :param topology parsed_topology: A topology already parsed to search for
     matches.
    :param str kind: The kind of object that posses the attribute to search for
     can be either 'nodes', 'ports' or 'links'.

    :return: A list of matching nodes,ports or links containing the attribute.
    """

    matched_objects = []
    regex_attribute, regex_value = regex.split('=')

    # Look for attribute name matching
    for group in parsed_topology[kind]:
        attributes = group['attributes']
        for key in attributes:
            if match(regex_attribute, key) and match(regex_value,
                                                     attributes[key]):

                # Retain order, and avoid adding repeated nodes
                for found_obj in group[kind]:
                    if found_obj not in matched_objects:
                        matched_objects.append(found_obj)

    return matched_objects


def _expand_ports(parsed_topology, ports_definitions):
    """
    Expands a list of port definitions into the matching port names.

    A port definition is a string that can match none, one or more ports
    (by using wildcards). It can be an expression for port name matching, or
    for matching all ports that have an specific attribute value. For example:

    ::

        'nodea:1'
        'hs*:1*'
        'attr=1'

    :param dict parsed_topology: A dictionary of a fully parsed topology.
    :param list ports_definitions: A list of port definitions.

    :return: A list of matching ports.
    """

    expanded_ports = []

    for port_definition in ports_definitions:

        # Check if definition is for attribute matching
        if type(port_definition) is not tuple:
            expanded_ports.extend(
                _match_by_attr(port_definition, parsed_topology, 'ports'))
            continue

        # The definition is not attribute matching, but name matching
        for port_group in parsed_topology['ports']:
            for port in port_group['ports']:
                if fnmatch(port[0], port_definition[0]) and fnmatch(
                        port[1],
                        port_definition[1]) and port not in expanded_ports:
                    expanded_ports.append(port)

    return expanded_ports


def _expand_links(parsed_topology, links_definitions):
    """
    Expands a list of links definitions into the matching link names.

    A link definition is a string that can match none, one or more nodes
    (by using wildcards). It can be an expression for link name matching, or
    for matching all links that have an specific attribute value. For example:

    ::

        'nodea:1 -- nodeb:1'
        'node*:1 -- nodeb:12*'
        'attr=2'

    :param dict parsed_topology: A dictionary of a fully parsed topology.
    :param list link_definitions: A list of link definitions.
    :return: A list of matching links.
    """

    expanded_links = []

    for link_definition in links_definitions:

        # Check if definition is for attribute matching
        if type(link_definition) is not tuple:
            expanded_links.extend(
                _match_by_attr(link_definition, parsed_topology, 'links'))
            continue

        # The definition is not attribute matching, but name matching
        for link_group in parsed_topology['links']:
            link = link_group['endpoints']
            if fnmatch(link[0][0], link_definition[0][0]) and fnmatch(
                    link[0][1], link_definition[0][1]) and fnmatch(
                        link[1][0], link_definition[1][0]) and fnmatch(
                            link[1][1], link_definition[1]
                            [1]) and link not in expanded_links:
                expanded_links.append(link)

    return expanded_links


__all__ = ['parse_attribute_injection']



# Copyright (C) 2026 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Test suite for find_topology_in_python function.
"""

from pathlib import Path


FILES = Path(__file__).parent / 'files'
SZNDIR = FILES / 'topologies'


def test_find_topology_in_python():
    """
    Test that find_topology_in_python correctly identifies the topology
    defined in a Python file.
    """
    from pyszn.parser import find_topology_in_python

    # Found in python file (TOPOLOGY)
    topology = find_topology_in_python(
        FILES / 'topodef.py',
        szn_dir=[SZNDIR],
    )
    assert topology is not None

    # Found in python file (TOPOLOGY_ID)
    topology_a = find_topology_in_python(
        FILES / 'topofile1.py',
        szn_dir=[SZNDIR],
    )
    assert topology_a is not None

    topology_b = find_topology_in_python(
        FILES / 'topofile2.py',
        szn_dir=[SZNDIR],
    )
    assert topology_b is not None

    # Topology not found
    topology_c = find_topology_in_python(
        FILES / 'topofile3.py',
        szn_dir=[SZNDIR],
    )
    assert topology_c is None

    # Bad Python file
    topology_d = find_topology_in_python(
        FILES / 'badfile.py',
        szn_dir=[SZNDIR],
    )
    assert topology_d is None
